"""
__version__ = "$Revision: 1.7 $"
__date__ = "$Date: 2001/08/18 18:41:30 $"
"""

import string
from error import *
from log import *

class Handler :
    """
    Maps a function name to a function, parsing the
    PythonCard-specific name into meaningful parts.
    """

    def __init__( self, aFunctionName, aFunctionObject ) :

        # Hold on to the full name.
        self._name = aFunctionName

        # Parse the function/method name.
        parts = string.split( aFunctionName, '_' )

        # The 'on' prefix, i.e. on_button1_click  
        self._prefix = parts[ 0 ] 

        # The name of the Widget instance that is the target.
        self._sourceName = parts[ 1 ]

        # The Event class identifier.
        self._eventName = parts[ 2 ]

        # Grab the function/method object.
        self._function = aFunctionObject
        
    def getName( self ) :
        return self._name 
        
    def getSourceName( self ) :
        return self._sourceName

    def getEventName( self ) :
        return self._eventName

    def getFunction( self ) :
        return self._function

    def __repr__( self ) :
        return str( self.__dict__ )

class ObjectLookup :
    """
    ObjectLookup - ABSTRACT CLASS

    Having a single event queue object centralizes the posting
    of events and makes it easier to abstract away wxWindows

    This would be even better if it followed the Singleton pattern
    where there could only be ONE EventQueue object in the system.

    Adding the Observer/Observable pattern makes it easier to write
    an event watcher that is decoupled from the EventQueue.

    See monitor.py and monitor_test.py

    Provides an interface for looking up objects by path.
    """

    # Brain dead way to implement the Singleton pattern
    # in Python.  This class CANNOT be subclassed, but then
    # again, we may not ever need to. If we do, we'll search
    # on Google for a better way to implement Singleton :).

    def __call__( self ) :
        return self

    def findByName( self, targetPath ) :
        # Must be overridden
        raise AbstractMethodException

    def findId( self, targetPath ) :
        # Must be overridden
        raise AbstractMethodException

class ChangeEvent :
    """
    A ChangeEvent indicates that some attribute of
    and object has changed.  The event carries a
    reference to the object, and the attribute
    value that changed.
    """

    def __init__( self, aObject, aOldValue ) :
        """
        Initialize this instance.
        """
        self._object = aObject
        self._oldValue = aOldValue

    def getChangedObject( self ) :
        """
        Get a reference to the object that was modified.
        """
        return self._object 

    def getOldValue( self ) :
        """
        Get the old attribute value that changed.
        NOTE: We'll probably need to provide a
        getAttributeName() method to identify the
        actual value that changed within the object.
        """
        return self._oldValue

class ChangeListener :
    """
    Defines an interface that must be implemented
    by classes that want to listen for ChangeEvents.
    """

    def changed( aChangeEvent ) :
        """
        Called by a Changeable object when it's
        contents are modified.
        """
        raise AbstractMethodException

class Changeable :
    """
    Defines an interface that must be implemented by
    classes that want to generate a notification when
    their contents change.
    """

    def __init__( self ) :
        """
        Initialize this instance.
        """
        self._listeners = []

    def addChangeEventListener( self, aChangeListener ) :
        """
        Add a ChangeListener to this Changeable object's
        list of interested listeners.
        """
        self._listeners.append( aChangeListener )

    def fireChanged( self, oldValue ) :
        """
        Broadcast a ChangeEvent to all registered ChangeListeners.
        """
        event = ChangeEvent( self, oldValue )
        for listener in self._listeners :
            listener.changed( event )

class Event :
    """
    Superclass of all event classes.
    """

    def __init__( self, aSource ) :
        self._source = aSource

    def getSource( self ) :
        """
        Get a reference to the object that
        generated this Event.
        """
        return self._source

    def getName( self ) :
        return EventMap().classToName( self.__class__ )

    def sendTo( self, aOwner, aHandler ) :
        """
        Tell an Event to dispatch itself to aOwner by
        calling aHandler.getFunction().
        """
        aHandler.getFunction()( aOwner, self.getSource(), self )

    def accept( self, aVisitor, aParam ) :
        """
        A generic implementation of the 'visitee' part
        of the Visitor pattern.
        """
        getattr( aVisitor, 'visit' + self.__class__.__name__ )( aParam )

    def getHandlerName( self ) :
        return 'on_' + self._source._getName() + '_' + self.getName()

# NOTE: The following two methods are only provided to help
# determine what event info should be transferred from wxEvents 
# to PythonCard Events.  They will go away some time soon.

    def setNativeEvent( self, aNativeEvent ) :
        self._nativeEvent = aNativeEvent

    def getNativeEvent( self ) :
        return self._nativeEvent

    def skip( self ) :
        self.getNativeEvent().Skip()


class WidgetEvent( Event ) :
    """
    A widget specific Event subclass.
    """

    def __init__( self, aWidgetName, aSource ) :
        Event.__init__( self, aSource )
        self._widgetName = widgetName
    
    def getWidgetName( self ) :
        return self._widgetName
    
class MessageEvent( Event ) :
    """
    A generic messaging event that accepts a target path
    and the method to call on the object identified by
    the target path.
    """

    def __init__( self, aTargetPath, aMessage, aParam, aSource ) :
        Event.__init__( self, aSource )
        self._targetPath = aTargetPath
        self._message = aMessage
        self._param = aParam

    def getTargetPath( self ) :
        return self._targetPath.split( '.' )

    def getMessage( self ) :
        return self._message

    def sendTo( self, aOwner, aHandler ) :
        """
        Tell an Event to dispatch itself to aOwner by
        calling aHandler.getFunction().
        """
        name = self.getTargetPath()[ 0 ]
        target = aOwner.name

        getattr( target, self.getMessage() )()


class EventSource :
    """
    All Components are EventSources.
    """

    def __init__( self ) :
        self._listeners = []
        
    def addEventListener( self, listener ) :
        """
        Add an EventListener as an observer of this object.
        """
        self._listeners.append( listener )
    
    def notifyEventListeners( self, event ) :
        """
        Notify all of our EventListeners that an event has occured.
        """
        for listener in self._listeners :
            listener.eventOccurred( event )

class EventListener :
    """
    ABSTRACT INTERFACE

    Define an interface that clients can implement in order
    to act as observers of the EventQueue.
    """

    # I don't know if it's possible to make a method
    # 'abstract' in python, i.e. specify that it MUST
    # be overridden?

    def eventOccurred( self, event ) :
        raise AbstractMethodException
               
class MenuEvent( Event ) :
    """
    The Event sent when an unbound Menu item is clicked.
    """

    def __init__( self, aMenuItemId, aMenu ) :
        Event.__init__( self, aMenu )
        self._itemId = aMenuItemId

    def getMenuItemId( self ) :
        return self._itemId

class CommandEvent( Event ) :
    """
    The Event sent when by a command source object
    like Button, ImageButton, MenuItem, etc.
    """

    def __init__( self, aName, aSource ) :
        Event.__init__( self, aSource )
        self._name = aName

    def getCommand( self  ) :
        return self._name
 
    def getHandlerName( self ) :
        return 'on_' + self.getCommand() + '_' + self.getName()

class GainFocusEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class LoseFocusEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class KeyPressedEvent( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


class KeyDownEvent( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


class KeyUpEvent( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


class TextUpdateEvent( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


class TextEnterEvent( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class SelectEvent( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


class IdleEvent( Event ) :
                           
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseClickEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseDoubleClickEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseDownEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseUpEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseMiddleDownEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseMiddleUpEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseMiddleDoubleClickEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseContextDownEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseContextUpEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseContextDoubleClickEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseEnterEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseLeaveEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseMoveEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )

class MouseDragEvent( Event ) :
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


MOUSE_EVENTS = [
                 GainFocusEvent,
                 LoseFocusEvent,
                 MouseDoubleClickEvent,
                 MouseDownEvent,
                 MouseUpEvent,
                 MouseMiddleDownEvent,
                 MouseMiddleUpEvent,
                 MouseMiddleDoubleClickEvent,
                 MouseContextDownEvent,
                 MouseContextUpEvent,
                 MouseContextDoubleClickEvent,
                 MouseEnterEvent,
                 MouseLeaveEvent,
                 MouseMoveEvent 
                 ]

# System Events

class CloseField( Event ) :
    
    def __init__( self, aSource ) :
        Event.__init__( self, aSource )


class EventMap :
    """
    Maps Event classes to handler names for use in user code, and vice-versa.
    """
    def __init__( self ) :
    
        self._table = [
            [ GainFocusEvent, 'gainFocus' ],
            [ LoseFocusEvent, 'loseFocus' ],
            [ MouseClickEvent, 'mouseClick' ],
            [ MouseDownEvent, 'mouseDown' ],
            [ MouseUpEvent, 'mouseUp' ],
            [ MouseDoubleClickEvent, 'mouseDoubleClick' ],
            [ MouseContextDownEvent, 'mouseContextDown' ],
            [ MouseContextUpEvent, 'mouseContextUp' ],
            [ MouseContextDoubleClickEvent, 'mouseContextDoubleClick' ],
            [ MouseMiddleDownEvent, 'mouseMiddleDown' ],
            [ MouseMiddleUpEvent, 'mouseMiddleUp' ],
            [ MouseMiddleDoubleClickEvent, 'mouseMiddleDoubleClick' ],
            [ MouseEnterEvent, 'mouseEnter' ],
            [ MouseLeaveEvent, 'mouseLeave' ],
            [ MouseMoveEvent, 'mouseMove' ],
            [ MouseDragEvent, 'mouseDrag' ],
            [ CommandEvent, 'command' ],
            [ KeyPressedEvent, 'keyPressed' ],
            [ KeyDownEvent, 'keyDown' ],
            [ KeyUpEvent, 'keyUp' ],
            [ TextUpdateEvent, 'textUpdate' ],
            [ TextEnterEvent, 'textEnter' ],
            [ SelectEvent, 'select' ],
            [ IdleEvent, 'idle' ],
            # System Events
            [ CloseField, 'closeField' ]
        ]
    
        self._classToNameMap = {}
    
        self._nameToClassMap = {}
    
        self._createClassToNameMap()
    
        self._createNameToClassMap()
    
    def _createClassToNameMap( self ) :
        for row in self._table :
            self._classToNameMap[ row[ 0 ] ] = row[ 1 ]
    
    def _createNameToClassMap( self ) :
        for row in self._table :
            self._nameToClassMap[ row[ 1 ] ] = row[ 0 ]
    
    def classToName( self, aEventClass ) :
        return self._classToNameMap[ aEventClass ]
    
    def nameToClass( self, aName ) :
        return self._nameToClassMap[ aName ]
    
    def __repr__( self ) :
        return str( self._classToNameMap ) + '\n\n' + str( self._nameToClassMap )
    
queue = None

class EventQueue :
    """
    Manages the dispatching of Events.

    The EventQueue class delegates to a single instance
    of EventQueueImpl - yansi ( yet another singleton implementation ;)
    """

    _impl = None

    class EventQueueImpl( EventSource, EventListener ) :

        def __init__( self ) :
            EventSource.__init__( self )

        def eventOccurred( self, aEvent ) :
            self.notifyEventListeners( aEvent )

    def __init__( self ) :
        if EventQueue._impl is None :
            EventQueue._impl = self.EventQueueImpl()

    def listenTo( self, aEventSource ) :
        aEventSource.addEventListener( self._impl )

    def addListener( self, aEventListener ) :
        self._impl.addEventListener( aEventListener )
    
# Unit Test

class Dummy :

    def visitIdleEvent( self, aParam ) :
        Log.getInstance().debug( 'visitIdleEvent(): ', aParam )

# Unit Test

if __name__ == '__main__' :

    print 'Unit Test - event'

    dummy = Dummy()

    event = IdleEvent( dummy )

    print event

    event.accept( dummy, 'this is a parameter' )


